import Foundation
import SpriteKit
import Connect4Core

class PieceNode : SKNode {
    static let pieceSize: CGFloat = 100
    static let borderFactor: CGFloat = 0.85
    
    private var gameScene: GameScene { self.scene as! GameScene }
    private var circle: SKShapeNode
    
    var onDragHandler: ((PieceNode) -> ())? = nil

    var piece: Piece = .init(withOwner: .noOne) {
        didSet {
            #if os(iOS)
            circle.fillColor = UIColor(piece.owner.pieceColor)
            #else
            circle.fillColor = NSColor(piece.owner.pieceColor)
            #endif
            circle.strokeColor = .pieceBorder
        }
    }
    var boardPosition: (x: Int, y: Int)? = nil {
        didSet {
            if let pos = boardPosition {
                position = CGPoint(
                    x: CGFloat(pos.x) * Self.pieceSize + Self.pieceSize / 2,
                    y: CGFloat(pos.y) * Self.pieceSize + Self.pieceSize / 2
                )
                zPosition = 10
            } else {
                zPosition = 100
            }
        }
    }
    
    override var isUserInteractionEnabled: Bool {
        get { self.onDragHandler != nil }
        set { /* fatalError("Unsupported operation: set") */ }
    }

    override init() {
        circle = SKShapeNode(circleOfRadius: Self.pieceSize / 2 * Self.borderFactor)

        super.init()
        
        addChild(circle)
        zPosition = 100
    }
    
    convenience init(onReleaseFn: @escaping (PieceNode) -> ()) {
        self.init()
        
        self.onDragHandler = onReleaseFn
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    public func clampedPosition() -> (row: Int, column: Int) {
        (
            row: Int((position.y) / Self.pieceSize),
            column: Int((position.x) / Self.pieceSize)
        )
    }
    
    private func dragTo(position pos: CGPoint) {
        self.position = pos
    }
    
    private func releaseAt(position pos: CGPoint) {
        self.dragTo(position: pos)
        onDragHandler?(self)
    }
    
#if os(macOS)
    override public func mouseDown(with event: NSEvent) {
        guard let holder = self.parent
        else { return }
        
        let pos = event.location(in: holder)
        self.dragTo(position: CGPoint(x: pos.x, y: pos.y))
    }
    
    override public func mouseDragged(with event: NSEvent) {
        guard let holder = self.parent
        else { return }
        
        let pos = event.location(in: holder)
        self.dragTo(position: CGPoint(x: pos.x, y: pos.y))
    }
    
    override public func mouseUp(with event: NSEvent) {
        guard let holder = self.parent
        else { return }
        
        let pos = event.location(in: holder)
        self.releaseAt(position: CGPoint(x: pos.x, y: pos.y))
    }
#elseif os(iOS)
    override public func touchesBegan(_ touches: Set<UITouch>, with event: UIEvent?) {
        guard
            let touch = touches.first,
            let holder = self.parent
        else { return }

        let pos = touch.location(in: holder)
        self.dragTo(position: CGPoint(x: pos.x, y: pos.y))
    }
    
    override public func touchesMoved(_ touches: Set<UITouch>, with event: UIEvent?) {
        guard
            let touch = touches.first,
            let holder = self.parent
        else { return }
        
        let pos = touch.location(in: holder)
        self.dragTo(position: CGPoint(x: pos.x, y: pos.y))
    }
    
    override public func touchesEnded(_ touches: Set<UITouch>, with event: UIEvent?) {
        guard
            let touch = touches.first,
            let holder = self.parent
        else { return }
        
        let pos = touch.location(in: holder)
        self.releaseAt(position: CGPoint(x: pos.x, y: pos.y))
    }
#endif
}
