import SwiftUI
import SpriteKit
import Connect4Core

struct IngameView: View {
    @StateObject
    private var vm: IngameVM
    
    var body: some View {
        VStack {
            HStack {
                PlayerView(
                    who: vm.player1,
                    dock_left: true,
                    isTurn: self.$vm.currentPlayer.map { $0 == .player1 }
                )
                Spacer()
                PlayerView(
                    who: vm.player2,
                    dock_left: false,
                    isTurn: self.$vm.currentPlayer.map { $0 == .player2 }
                )
            }
            
            let statusText: LocalizedStringKey = switch (self.vm.result) {
            case .even: "inGame.status.result.even"
            case .winner(let winner, _):
                switch (winner) {
                case .player1: "inGame.status.result.winner \(self.vm.player1.name)"
                case .player2: "inGame.status.result.winner \(self.vm.player2.name)"
                default: fatalError("Unreachable")
                }
            default:
                switch (self.vm.currentPlayer) {
                case .player1: "inGame.status.turn \(self.vm.player1.name)"
                case .player2: "inGame.status.turn \(self.vm.player2.name)"
                default: "inGame.status.new"
                }
            }
            
            Text(statusText).tint(.accentColor)

            Spacer()
            
            VStack(alignment: .center) {
                SpriteView(scene: vm.scene, options: .allowsTransparency)
                    .aspectRatio(
                        vm.scene.size.width / vm.scene.size.height,
                        contentMode: .fit
                    )
            }.safeAreaPadding(.horizontal)
            
            Spacer()
            
            // let style: (key: LocalizedStringKey, img: String) = if vm.paused {
            //     ("inGame.ppButton.resume", "play.circle")
            // } else {
            //     ("inGame.ppButton.pause", "pause.circle")
            // }
            // Button(style.key, systemImage: style.img) {
            //     vm.paused.toggle()
            // }
            // .`if`(vm.paused, { $0.buttonStyle(.bordered) })
            // .`if`(!vm.paused, { $0.buttonStyle(.borderedProminent) })
            // //For some reason the following crashes the compiler:
            // // .either(vm.paused,
            // //         `true`: { $0.buttonStyle(.borderedProminent) },
            // //         `false`: { $0.buttonStyle(.bordered) }
            // // )
            // .tint(.accentColor)
            //
            // Spacer()

            Text("inGame.currentRules \(vm.rulesName)")
        }.onAppear {
            vm.start()
        }
    }
    
    init(vm: IngameVM) {
        self._vm = StateObject(wrappedValue: vm)
    }
}

private struct PlayerView: View {
    let dockLeft: Bool
    let who: PlayerVM
    
    @Binding
    var isTurn: Bool

    var body: some View {
        let textAlignment: HorizontalAlignment = if self.dockLeft {
            .leading
        } else {
            .trailing
        }
        
        VStack {
            HStack {
                let img = Circle().frame(width: 50, height: 50)
                
                if self.dockLeft {
                    img
                }
                
                VStack(alignment: textAlignment) {
                    Text(who.name)
                    
                    let type = String(localized: String.LocalizationValue(who.type.baseTranslationKey))
                    
                    if type != who.name {
                        Text(type)
                    }
                }
                
                if !self.dockLeft {
                    img
                }
            }
            
            HStack {
                let wheel = ProgressView()
                    .progressViewStyle(.circular)
                    .tint(.accentColor)
                    .transition(.scale)
                    .if(!self.isTurn, { $0.hidden() })
                
                if !self.dockLeft {
                    wheel
                }
                
                // TODO
                // let time = "??:??:??"
                // Text(time).padding(.horizontal, 10)
                
                if self.dockLeft {
                    wheel
                }
            }
        }.padding(.all, 5)
            .border(who.id.pieceColor)
            .containerShape(RoundedRectangle(cornerRadius: 5))
            .padding(.all, 5)
    }
    
    init(who owner: PlayerVM, dock_left dockLeft: Bool, isTurn: Binding<Bool>) {
        self.who = owner
        self.dockLeft = dockLeft
        self._isTurn = isTurn
    }
}

#Preview {
    IngameView(vm: IngameVM(settings: NewGameVM(), player1: PlayerSettingsVM(type: .Human), player2: PlayerSettingsVM(type: .AISimpleNegaMax)))
}
