import Foundation
import Connect4Core
import Connect4Players
import Connect4Rules
import Connect4Persistance

class IngameVM: ObservableObject {
    let gameName: String
    
    let game: Game
    
    /// This field is a "fix" for a random double initialization of the VM
    /// causing a desync (how?) between the VM of the view and the VM of the
    /// scene.
    ///
    /// The real issue is the double initialization but I suspect it is caused
    /// by a TextField loosing focus and triggering a change and recomposition.
    var scene: GameScene! = nil

    // @Published
    var rulesName: String { game.rules.name }
    
    var columns: Int { game.rules.nbColumns }
    /// alias for columns
    var width: Int { self.columns }
    var rows: Int { game.rules.nbRows }
    /// alias for rows
    var height: Int { self.rows }
    
    var player1: PlayerVM
    var player2: PlayerVM
    
    @Published
    var currentPlayer: Owner = .noOne
    
    @Published
    var currentBoard: Board
    
    @Published
    var result: Connect4Core.Result = .notFinished
    
    private var running: Bool = false
    
    init(game: Game) {
        let fmt = DateFormatter()
        fmt.locale = Locale(identifier: "EN") // No "root" locale?
        self.gameName = fmt.string(from: Date.now)
        
        self.game = game
        
        self.player1 = PlayerVM(inner: game.players[.player1]!)
        self.player2 = PlayerVM(inner: game.players[.player2]!)
        
        self.currentBoard = self.game.board
        
        self.scene = GameScene(viewModel: self)
        
        game.addGameOverListener { board, result, player in
            DispatchQueue.main.async {
                self.currentPlayer = .noOne
                self.result = result
            }
        }
        game.addGameChangedListener { game, result in
            if game.players.contains(where: { $0.value is ReplayPlayer }) {
                // Do not save progress of an already finished game
                return
            }
            
            defer {
                print("game saved")
            }
            
            if result == .notFinished {
                _ = try await Persistance.saveGame(
                    withName: Persistance.unfinishedGameFileName,
                    andGame: self.game,
                    withFolderName: Persistance.saveDirectory
                )
                return
            }
            
            try await Persistance.deleteGame(
                withName: Persistance.unfinishedGameFileName,
                withFolderName: Persistance.saveDirectory
            )
            
            _ = try await Persistance.saveGameResult(
                withName: Persistance.finishedGameFileName,
                andGame: game,
                andResult: result,
                withFolderName: Persistance.saveDirectory
            )
        }
        game.addBoardChangedListener { board, lastCell in
            DispatchQueue.main.async {
                self.currentBoard = board
            }
        }
        game.addPlayerNotifiedListener { board, player in
            DispatchQueue.main.async {
                self.currentPlayer = player.id
            }
        }
    }

    convenience init(settings: NewGameVM, player1: PlayerSettingsVM, player2: PlayerSettingsVM) {
        let rulesInit: (Int, Int, Int) -> (any Rules)? = switch (settings.rulesType) {
        case .Classic: Connect4Rules.init
        case .TicTacToe: TicTacToeRules.init
        case .PopOut: PopOutRules.init
        }
        let rules = rulesInit(
            Int(settings.height),
            Int(settings.width),
            Int(settings.alignedTokens)
        )!
        
        let player1 = Self.playerOf(settings: player1, id: .player1)!
        let player2 = Self.playerOf(settings: player2, id: .player2)!
        
        let game = try! Game(withRules: rules, andPlayer1: player1, andPlayer2: player2)
        
        self.init(game: game)
    }
    
    public func start() {
        if running { return }
        running = true
        Task.detached(priority: .userInitiated) {
            try await self.game.start()
        }
    }
    
    public func pieceDropped(at: (row: Int, column: Int)) {
        Task.detached(priority: .userInitiated) {
            try await self.game.onPlayed(move: Move(
                of: self.currentPlayer,
                toRow: at.row,
                toColumn: at.column
            ))
        }
    }
    
    private static func playerOf(settings: PlayerSettingsVM, id: Owner) -> Player? {
        let name = settings.name.nilIfEmpty ?? String(localized: String.LocalizationValue(settings.type.baseTranslationKey)) 
        
        return switch (settings.type) {
        case .Human:
            HumanPlayer(withName: name, andId: id)
        case .AIRandom:
            RandomPlayer(withName: name, andId: id)
        case .AIFinnishHim:
            FinnishHimPlayer(withName: name, andId: id)
        case .AISimpleNegaMax:
            SimpleNegaMaxPlayer(withName: name, andId: id)
        }
    }
}
