import Foundation
import Connect4Core
import Connect4Persistance

class SavedGamesVM: ObservableObject {
    @Published
    var unfinished: [GameEntryVM] = []
    @Published
    var finished: [GameEntryVM] = []
    
    init() {
        Task(priority: .userInitiated) {
            let sortedU: [GameEntryVM]
            let sortedF: [GameEntryVM]
            var temp: [GameEntryVM]

            do {
                if let game = try await Persistance.loadGame(withName: Persistance.unfinishedGameFileName, withFolderName: Persistance.saveDirectory) {
                    temp = [GameEntryVM(fromGame: game)]
                } else {
                    temp = []
                }
                sortedU = temp.sorted(by: { $0.date > $1.date })
            } catch {
                sortedU = []
            }
            
            do {
                let results = (try await Persistance.loadGameResults(withName: Persistance.finishedGameFileName, withFolderName: Persistance.saveDirectory)) ?? []
                temp = results.map(GameEntryVM.init)
                sortedF = temp.sorted(by: { $0.date > $1.date })
            } catch {
                sortedF = []
            }
            
            DispatchQueue.main.async {
                self.unfinished = sortedU
                self.finished = sortedF
            }
        }
    }
}

class GameEntryVM: Identifiable {
    var id: Date { self.date }
    
    let game: Game?
    
    let date: Date
    let player1Name: String?
    let player1Type: PlayerType
    let player2Name: String?
    let player2Type: PlayerType
    let rules: RulesType
    
    init(fromResult from: GameResult) {
        self.date = from.date
        let (p1, p2) = if from.players[0].id == .player1 {
            (from.players[0], from.players[1])
        } else {
            (from.players[1], from.players[0])
        }
        
        self.player1Type = Self.i(data: p1)
        self.player1Name = p1.name.nilIfEmpty
        self.player2Type = Self.i(data: p2)
        self.player2Name = p2.name.nilIfEmpty

        self.rules = switch (from.rules.type) {
        case "Connect4Rules": .Classic
        case "TicTacToeRules": .TicTacToe
        case "PopOutRules": .PopOut
        default: fatalError("Unexpected rules type")
        }
        
        self.game = nil
    }
    
    init(fromGame from: Game) {
        self.game = from;
        self.date = Date.now
        
        let p1 = from.players[.player1]!
        let p2 = from.players[.player2]!
        
        self.player1Name = p1.name.nilIfEmpty
        self.player1Type = p1.type
        self.player2Name = p2.name.nilIfEmpty
        self.player2Type = p2.type

        self.rules = from.rules.type
    }
    
    private static func i(data: PlayerData) -> PlayerType {
        return switch (data.type) {
        case "HumanPlayer": .Human
        case "RandomPlayer": .AIRandom
        case "FinnishHimPlayer": .AIFinnishHim
        case "SimpleNegaMaxPlayer": .AISimpleNegaMax
        default: fatalError("Unexpected player type")
        }
    }
}
